from gymnax.utils.state_translate import np_state_to_jax
from gymnax.utils.test_helpers import (
    assert_correct_state,
    assert_correct_transit,
    minatar_action_map,
)


__all__ = [
    "np_state_to_jax",
    "assert_correct_state",
    "assert_correct_transit",
    "minatar_action_map",
]



import jax
import jaxlib
import numpy as np
from .state_translate import np_state_to_jax


def assert_correct_state(
    env_gym, env_name: str, state_jax: dict, atol: float = 1e-4
):
    """Check that numpy-based env state is same as JAX dict."""
    state_gym = np_state_to_jax(env_gym, env_name)
    # print(state_gym)
    # Loop over keys and assert that individual entries are same/close
    for k in state_gym.keys():
        jax_value = getattr(state_jax, k)
        # print(k)
        if k not in ["time", "terminal"]:
            if type(jax_value) in [
                jax.interpreters.xla._DeviceArray,
                jaxlib.xla_extension.Buffer,
                np.ndarray,
            ]:
                assert np.allclose(jax_value, state_gym[k], atol=atol)
            else:
                # print(k, state_gym[k], state_jax[k])
                # Exclude extra time and terminal state from assertion
                if type(state_gym[k]) in [float, np.float64]:
                    np.allclose(state_gym[k], jax_value, atol=atol)
                else:
                    assert state_gym[k] == jax_value


def assert_correct_transit(
    obs_gym,
    reward_gym,
    done_gym,
    obs_jax,
    reward_jax,
    done_jax,
    atol: float = 1e-4,
):
    """Check that obs, reward, done transition outputs are correct."""
    if not done_gym:
        assert np.allclose(obs_gym, obs_jax, atol=atol)
    assert np.allclose(reward_gym, reward_jax, atol=atol)
    assert np.alltrue(done_gym == done_jax)


def minatar_action_map(action_jax: int, env_name: str):
    """Helper that maps gymnax MinAtar action to the numpy equivalent."""
    all_actions = ["n", "l", "u", "r", "d", "f"]
    if env_name == "Asterix-MinAtar":
        minimal_actions = ["n", "l", "u", "r", "d"]
    elif env_name == "Breakout-MinAtar":
        minimal_actions = ["n", "l", "r"]
    elif env_name == "Freeway-MinAtar":
        minimal_actions = ["n", "u", "d"]
    elif env_name == "Seaquest-MinAtar":
        minimal_actions = ["n", "l", "u", "r", "d", "f"]
    elif env_name == "SpaceInvaders-MinAtar":
        minimal_actions = ["n", "l", "r", "f"]
    else:
        raise ValueError(f"{env_name} not in implemented MinAtar environments.")
    action_idx = all_actions.index(minimal_actions[action_jax])
    return action_idx



import matplotlib.pyplot as plt
import matplotlib.animation as animation
from typing import Optional
from .vis_minatar import init_minatar, update_minatar
from .vis_circle import init_circle, update_circle
from .vis_maze import init_maze, update_maze


class Visualizer(object):
    def __init__(self, env, env_params, state_seq: list):
        self.env = env
        self.env_params = env_params
        self.state_seq = state_seq
        self.fig, self.ax = plt.subplots(1, 1, figsize=(6, 5))
        self.interval = 500

    def animate(
        self,
        save_fname: Optional[str] = "test.gif",
        view: bool = False,
    ):
        """Anim for 2D fct - x (#steps, #pop, 2) & fitness (#steps, #pop)"""
        ani = animation.FuncAnimation(
            self.fig,
            self.update,
            frames=len(self.state_seq),
            init_func=self.init,
            blit=False,
            interval=self.interval,
        )
        # Save the animation to a gif
        if save_fname is not None:
            ani.save(save_fname)
        # Simply view it 3 times
        if view:
            plt.show(block=False)
            plt.pause(3)
            plt.close()

    def init(self):
        # Plot placeholder points
        if self.env.name in [
            "Asterix-MinAtar",
            "Breakout-MinAtar",
            "Freeway-MinAtar",
            "Seaquest-MinAtar",
            "SpaceInvaders-MinAtar",
        ]:
            self.im = init_minatar(self.ax, self.env, self.state_seq[0])
        elif self.env.name == "PointRobot-misc":
            self.im = init_circle(
                self.ax, self.env, self.state_seq[0], self.env_params
            )
        elif self.env.name == "MetaMaze-misc":
            self.im = init_maze(
                self.ax, self.env, self.state_seq[0], self.env_params
            )
        self.fig.tight_layout(rect=[0.02, 0.03, 1.0, 0.95])

    def update(self, frame):
        if self.env.name in [
            "Asterix-MinAtar",
            "Breakout-MinAtar",
            "Freeway-MinAtar",
            "Seaquest-MinAtar",
            "SpaceInvaders-MinAtar",
        ]:
            update_minatar(self.im, self.env, self.state_seq[frame])
        elif self.env.name == "PointRobot-misc":
            self.im = update_circle(self.im, self.env, self.state_seq[frame])
        elif self.env.name == "MetaMaze-misc":
            self.im = update_maze(self.im, self.env, self.state_seq[frame])
        self.ax.set_title(f"{self.env.name} - Time Step {frame + 1}")



import numpy as np


def init_maze(ax, env, state, params):
    im = ax.imshow(env.env_map)
    anno_pos = ax.annotate(
        "A",
        fontsize=20,
        xy=(state.pos[1], state.pos[0]),
        xycoords="data",
        xytext=(state.pos[1] - 0.3, state.pos[0] + 0.25),
    )
    anno_goal = ax.annotate(
        "G",
        fontsize=20,
        xy=(state.goal[1], state.goal[0]),
        xycoords="data",
        xytext=(state.goal[1] - 0.3, state.goal[0] + 0.25),
    )
    ax.set_xticks([])
    ax.set_yticks([])
    return anno_pos


def update_maze(im, env, state):
    xy = (state.pos[1], state.pos[0])

    xytext = (state.pos[1] - 0.3, state.pos[0] + 0.25)

    im.set_position((xytext[0], xytext[1]))
    im.xy = (xy[0], xy[1])
    return im



import numpy as np


def init_circle(ax, env, state, params):
    import matplotlib.pyplot as plt

    angles = np.linspace(0, np.pi, 100)
    x, y = np.cos(angles), np.sin(angles)
    ax.plot(x, y, color="k")
    plt.axis("scaled")
    ax.set_xlim(-1.25, 1.25)
    ax.set_ylim(-0.25, 1.25)
    ax.set_xticks([])
    ax.set_yticks([])

    anno_goal = plt.Circle(
        (state.goal[0], state.goal[1]), radius=params.goal_radius, alpha=0.3
    )
    ax.add_artist(anno_goal)

    anno_agent = plt.Circle(
        (state.pos[0], state.pos[1]), radius=0.05, alpha=1, color="red"
    )
    ax.add_artist(anno_agent)
    return [anno_goal, anno_agent]


def update_circle(im, env, state):
    anno_goal = im[0]
    anno_agent = im[1]
    anno_goal.center = (state.goal[0], state.goal[1])
    anno_agent.center = (state.pos[0], state.pos[1])
    return [anno_goal, anno_agent]



from .environments.classic_control import (
    Pendulum,
    CartPole,
    MountainCar,
    ContinuousMountainCar,
    Acrobot,
)

# =============================================================================
from .environments.bsuite import (
    Catch,
    DeepSea,
    DiscountingChain,
    MemoryChain,
    UmbrellaChain,
    MNISTBandit,
    SimpleBandit,
)

# =============================================================================
from .environments.minatar import (
    MinAsterix,
    MinBreakout,
    MinFreeway,
    MinSeaquest,
    MinSpaceInvaders,
)

# =============================================================================
from .environments.misc import (
    BernoulliBandit,
    GaussianBandit,
    FourRooms,
    MetaMaze,
    PointRobot,
)

# =============================================================================


def make(env_id: str, **env_kwargs):
    """A JAX-version of OpenAI's infamous env.make(env_name)"""
    # 1. Classic OpenAI Control Tasks
    if env_id == "Pendulum-v1":
        env = Pendulum(**env_kwargs)
    elif env_id == "CartPole-v1":
        env = CartPole(**env_kwargs)
    elif env_id == "MountainCar-v0":
        env = MountainCar(**env_kwargs)
    elif env_id == "MountainCarContinuous-v0":
        env = ContinuousMountainCar(**env_kwargs)
    elif env_id == "Acrobot-v1":
        env = Acrobot(**env_kwargs)

    # 2. DeepMind's bsuite environments
    elif env_id == "Catch-bsuite":
        env = Catch(**env_kwargs)
    elif env_id == "DeepSea-bsuite":
        env = DeepSea(**env_kwargs)
    elif env_id == "DiscountingChain-bsuite":
        env = DiscountingChain(**env_kwargs)
    elif env_id == "MemoryChain-bsuite":
        env = MemoryChain(**env_kwargs)
    elif env_id == "UmbrellaChain-bsuite":
        env = UmbrellaChain(**env_kwargs)
    elif env_id == "MNISTBandit-bsuite":
        env = MNISTBandit(**env_kwargs)
    elif env_id == "SimpleBandit-bsuite":
        env = SimpleBandit(**env_kwargs)

    # 3. MinAtar Environments
    elif env_id == "Asterix-MinAtar":
        env = MinAsterix(**env_kwargs)
    elif env_id == "Breakout-MinAtar":
        env = MinBreakout(**env_kwargs)
    elif env_id == "Freeway-MinAtar":
        env = MinFreeway(**env_kwargs)
    elif env_id == "Seaquest-MinAtar":
        env = MinSeaquest(**env_kwargs)
    elif env_id == "SpaceInvaders-MinAtar":
        env = MinSpaceInvaders(**env_kwargs)

    # 4. Miscellanoues Environments
    elif env_id == "BernoulliBandit-misc":
        env = BernoulliBandit(**env_kwargs)
    elif env_id == "GaussianBandit-misc":
        env = GaussianBandit(**env_kwargs)
    elif env_id == "FourRooms-misc":
        env = FourRooms(**env_kwargs)
    elif env_id == "MetaMaze-misc":
        env = MetaMaze(**env_kwargs)
    elif env_id == "PointRobot-misc":
        env = PointRobot(**env_kwargs)
    else:
        raise ValueError("Environment ID is not registered.")

    # Create a jax PRNG key for random seed control
    return env, env.default_params



# python3
# pylint: disable=g-bad-file-header
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Downloads and loads the MNIST dataset.

Adapted from https://github.com/google/jax/blob/master/examples/datasets.py
and https://github.com/deepmind/bsuite/blob/master/bsuite/utils/datasets.py
"""

import array
import gzip
import os
from os import path
import struct

from absl import logging
import numpy as np
from six.moves.urllib.request import urlretrieve


def _download(url, filename, directory="/tmp/mnist"):
    """Download a url to a file in the given directory."""
    if not path.exists(directory):
        os.makedirs(directory)
    out_file = path.join(directory, filename)
    if not path.isfile(out_file):
        urlretrieve(url, out_file)
        logging.info("Downloaded %s to %s", url, directory)


def load_mnist(directory="/tmp/mnist"):
    """Download and parse the raw MNIST dataset."""
    # CVDF mirror of http://yann.lecun.com/exdb/mnist/
    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"

    def parse_labels(filename):
        with gzip.open(filename, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            return np.array(array.array("B", fh.read()), dtype=np.uint8)

    def parse_images(filename):
        with gzip.open(filename, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            return np.array(array.array("B", fh.read()), dtype=np.int8).reshape(
                (num_data, rows, cols)
            )

    for filename in [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]:
        _download(base_url + filename, filename, directory)

    train_images = parse_images(path.join(directory, "train-images-idx3-ubyte.gz"))
    train_labels = parse_labels(path.join(directory, "train-labels-idx1-ubyte.gz"))
    test_images = parse_images(path.join(directory, "t10k-images-idx3-ubyte.gz"))
    test_labels = parse_labels(path.join(directory, "t10k-labels-idx1-ubyte.gz"))

    return (train_images, train_labels), (test_images, test_labels)



import numpy as np
import jax.numpy as jnp


def np_state_to_jax(env, env_name: str = "Pendulum-v1", get_jax: bool = False):
    """Helper that collects env state into dict for JAX `step`."""
    if env_name in [
        "Pendulum-v1",
        "CartPole-v1",
        "MountainCar-v0",
        "MountainCarContinuous-v0",
        "Acrobot-v1",
    ]:
        state_gym_to_jax = control_np_to_jax(env, env_name, get_jax)
    elif env_name in [
        "Catch-bsuite",
        "DeepSea-bsuite",
        "DiscountingChain-bsuite",
        "MemoryChain-bsuite",
        "UmbrellaChain-bsuite",
        "MNISTBandit-bsuite",
        "SimpleBandit-bsuite",
    ]:
        state_gym_to_jax = bsuite_np_to_jax(env, env_name, get_jax)
    elif env_name in [
        "Asterix-MinAtar",
        "Breakout-MinAtar",
        "Freeway-MinAtar",
        "Seaquest-MinAtar",
        "SpaceInvaders-MinAtar",
    ]:
        state_gym_to_jax = minatar_np_to_jax(env, env_name, get_jax)
    # TODO: Add misc/meta-learning environment testing
    else:
        raise ValueError(
            f"{env_name} is not in set of implemented environments."
        )
    return state_gym_to_jax


def control_np_to_jax(
    env, env_name: str = "Pendulum-v1", get_jax: bool = False
):
    """Collects env state of classic_control into dict for JAX `step`."""
    if env_name == "Pendulum-v1":
        state_gym_to_jax = {
            "theta": env.state[0],
            "theta_dot": env.state[1],
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.classic_control.pendulum import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "CartPole-v1":
        state_gym_to_jax = {
            "x": env.state[0],
            "x_dot": env.state[1],
            "theta": env.state[2],
            "theta_dot": env.state[3],
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.classic_control.cartpole import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "MountainCar-v0":
        state_gym_to_jax = {
            "position": env.state[0],
            "velocity": env.state[1],
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.classic_control.mountain_car import (
                EnvState,
            )

            return EnvState(**state_gym_to_jax)
    elif env_name == "MountainCarContinuous-v0":
        state_gym_to_jax = {
            "position": env.state[0],
            "velocity": env.state[1],
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.classic_control.continuous_mountain_car import (
                EnvState,
            )

            return EnvState(**state_gym_to_jax)
    elif env_name == "Acrobot-v1":
        state_gym_to_jax = {
            "joint_angle1": env.state[0],
            "joint_angle2": env.state[1],
            "velocity_1": env.state[2],
            "velocity_2": env.state[3],
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.classic_control.acrobot import EnvState

            return EnvState(**state_gym_to_jax)
    return state_gym_to_jax


def bsuite_np_to_jax(
    env, env_name: str = "Catch-bsuite", get_jax: bool = False
):
    """Collects env state of bsuite into dict for JAX `step`."""
    if env_name == "Catch-bsuite":
        state_gym_to_jax = {
            "ball_x": env._ball_x,
            "ball_y": env._ball_y,
            "paddle_x": env._paddle_x,
            "paddle_y": env._paddle_y,
            "prev_done": env._reset_next_step,
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.bsuite.catch import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "DeepSea-bsuite":
        state_gym_to_jax = {
            "row": env._row,
            "column": env._column,
            "bad_episode": env._bad_episode,
            "total_bad_episodes": env._total_bad_episodes,
            "denoised_return": env._denoised_return,
            "optimal_return": env._optimal_return,
            "action_mapping": env._action_mapping,
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.bsuite.deep_sea import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "DiscountingChain-bsuite":
        state_gym_to_jax = {
            "rewards": env._rewards,
            "context": env._context,
            "time": env._timestep,
        }
        if get_jax:
            from gymnax.environments.bsuite.discounting_chain import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "MemoryChain-bsuite":
        state_gym_to_jax = {
            "context": env._context,
            "query": env._query,
            "total_perfect": env._total_perfect,
            "total_regret": env._total_regret,
            "time": env._timestep,
        }
        if get_jax:
            from gymnax.environments.bsuite.memory_chain import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "UmbrellaChain-bsuite":
        state_gym_to_jax = {
            "need_umbrella": env._need_umbrella,
            "has_umbrella": env._has_umbrella,
            "total_regret": env._total_regret,
            "time": env._timestep,
        }
        if get_jax:
            from gymnax.environments.bsuite.umbrella_chain import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "MNISTBandit-bsuite":
        state_gym_to_jax = {
            "correct_label": env._correct_label,
            "regret": env._total_regret,
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.bsuite.mnist import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "SimpleBandit-bsuite":
        state_gym_to_jax = {
            "rewards": env._rewards,
            "total_regret": env._total_regret,
            "time": 0,
        }
        if get_jax:
            from gymnax.environments.bsuite.bandit import EnvState

            return EnvState(**state_gym_to_jax)
    return state_gym_to_jax


def minatar_np_to_jax(
    env, env_name: str = "Asterix-MinAtar", get_jax: bool = False
):
    """Collects env state of MinAtar into dict for JAX `step`."""
    if env_name == "Asterix-MinAtar":
        entities_array = jnp.zeros((8, 5), dtype=jnp.int32)
        for i in range(8):
            if env.env.entities[i] is not None:
                entities_array = entities_array.at[i, 0:4].set(
                    env.env.entities[i]
                )
                entities_array = entities_array.at[i, 4].set(1)
        state_gym_to_jax = {
            "player_x": env.env.player_x,
            "player_y": env.env.player_y,
            "shot_timer": env.env.shot_timer,
            "spawn_speed": env.env.spawn_speed,
            "spawn_timer": env.env.spawn_timer,
            "move_speed": env.env.move_speed,
            "move_timer": env.env.move_timer,
            "ramp_timer": env.env.ramp_timer,
            "ramp_index": env.env.ramp_index,
            "entities": entities_array,
            "time": 0,
            "terminal": 0,
        }
        if get_jax:
            from gymnax.environments.minatar.asterix import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "Breakout-MinAtar":
        state_gym_to_jax = {
            "ball_y": jnp.array(env.env.ball_y),
            "ball_x": jnp.array(env.env.ball_x),
            "ball_dir": env.env.ball_dir,
            "pos": env.env.pos,
            "brick_map": jnp.array(env.env.brick_map),
            "strike": env.env.strike,
            "last_y": jnp.array(env.env.last_y),
            "last_x": jnp.array(env.env.last_x),
            "time": 0,
            "terminal": 0,
        }
        if get_jax:
            from gymnax.environments.minatar.breakout import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "Freeway-MinAtar":
        state_gym_to_jax = {
            "pos": env.env.pos,
            "cars": jnp.array(env.env.cars),
            "move_timer": env.env.move_timer,
            "time": 0,
            "terminal": 0,
        }
        if get_jax:
            from gymnax.environments.minatar.freeway import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "Seaquest-MinAtar":
        f_bullets = np.zeros((100, 3))
        for i, f_b in enumerate(env.env.f_bullets):
            f_bullets[i] = f_b
        e_bullets = np.zeros((100, 3))
        for i, e_b in enumerate(env.env.e_bullets):
            e_bullets[i] = e_b
        e_fish = np.zeros((100, 5))
        for i, e_f in enumerate(env.env.e_fish):
            e_fish[i] = e_f + [10]
        e_subs = np.zeros((100, 5))
        for i, e_s in enumerate(env.env.e_subs):
            e_subs[i] = e_s
        divers = np.zeros((100, 4))
        for i, d in enumerate(env.env.divers):
            divers[i] = d

        state_gym_to_jax = {
            "oxygen": env.env.oxygen,
            "sub_x": env.env.sub_x,
            "sub_y": env.env.sub_y,
            "sub_or": env.env.sub_or,
            "f_bullet_count": len(env.env.f_bullets),
            "f_bullets": f_bullets,
            "e_bullet_count": len(env.env.e_bullets),
            "e_bullets": e_bullets,
            "e_fish_count": len(env.env.e_fish),
            "e_fish": e_fish,
            "e_subs_count": len(env.env.e_subs),
            "e_subs": e_subs,
            "diver_count": env.env.diver_count,
            "divers": divers,
            "e_spawn_speed": env.env.e_spawn_speed,
            "e_spawn_timer": env.env.e_spawn_timer,
            "d_spawn_timer": env.env.d_spawn_timer,
            "move_speed": env.env.move_speed,
            "ramp_index": env.env.ramp_index,
            "shot_timer": env.env.shot_timer,
            "surface": env.env.surface,
            "time": 0,
            "terminal": 0,
        }
        if get_jax:
            from gymnax.environments.minatar.seaquest import EnvState

            return EnvState(**state_gym_to_jax)
    elif env_name == "SpaceInvaders-MinAtar":
        state_gym_to_jax = {
            "pos": env.env.pos,
            "f_bullet_map": jnp.array(env.env.f_bullet_map),
            "e_bullet_map": jnp.array(env.env.e_bullet_map),
            "alien_map": jnp.array(env.env.alien_map),
            "alien_dir": env.env.alien_dir,
            "enemy_move_interval": env.env.enemy_move_interval,
            "alien_move_timer": env.env.alien_move_timer,
            "alien_shot_timer": env.env.alien_shot_timer,
            "ramp_index": env.env.ramp_index,
            "shot_timer": env.env.shot_timer,
            "ramping": env.env.ramping,
            "time": 0,
            "terminal": 0,
        }
        if get_jax:
            from gymnax.environments.minatar.space_invaders import EnvState

            return EnvState(**state_gym_to_jax)
    return state_gym_to_jax



import numpy as np


def init_minatar(ax, env, state):
    import seaborn as sns
    import matplotlib.colors as colors

    obs = env.get_obs(state)
    n_channels = env.obs_shape[-1]
    # The seaborn color_palette cubhelix is used to assign visually distinct colors to each channel for the env
    cmap = sns.color_palette("cubehelix", n_channels)
    cmap.insert(0, (0, 0, 0))
    cmap = colors.ListedColormap(cmap)
    bounds = [i for i in range(n_channels + 2)]
    norm = colors.BoundaryNorm(bounds, n_channels + 1)
    numerical_state = (
        np.amax(obs * np.reshape(np.arange(n_channels) + 1, (1, 1, -1)), 2)
        + 0.5
    )
    ax.set_xticks([])
    ax.set_yticks([])
    return ax.imshow(
        numerical_state, cmap=cmap, norm=norm, interpolation="none"
    )


def update_minatar(im, env, state):
    obs = env.get_obs(state)
    n_channels = env.obs_shape[-1]
    numerical_state = (
        np.amax(obs * np.reshape(np.arange(n_channels) + 1, (1, 1, -1)), 2)
        + 0.5
    )
    im.set_data(numerical_state)
